// ----------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "svgalib.cc", created on 17/04/99
 * descr : svgalib port specific code
 */
// ----------------------------------------------------------------------------

#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
#include <sys/mman.h>
#include <unistd.h>

#include "z11.h"
#include "Gfx.h"
#include "Input.h"
#include "EvKbd.h"
#include "FileRead.h"

FILEINFO(svgalib);

// ----------------------------------------------------------------------------
// GFX code
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
// sets one color register
void ScreenPalette::set(int i, RGB r) {
  gl_setpalettecolor(i, r.r>>2, r.g>>2, r.b>>2);
  data[i]=r;
}
// ----------------------------------------------------------------------------
// sets entire palette
ScreenPalette & ScreenPalette::operator =(ColourPalette & p) {
  memcpy(data,p.getData(),sizeof(data));
  byte * ptr = &data[0].r;
  for(int i=0; i<256*3; i++) *ptr++>>=2;
  gl_setpalette(data);
  memcpy(data,p.getData(),sizeof(data));
  return *this;
} 
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// bitmap:
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
// screen copy primitives
void Bmp::put(int x, int y) {
  Check(buffer); 
  gl_putbox(x,y,w,h,buffer); 
}
// ----------------------------------------------------------------------------
void Bmp::get(int x, int y) {
  Check(buffer); 
  gl_getbox(x,y,w,h,buffer); 
}
// ----------------------------------------------------------------------------
void Bmp::putS(int x, int y) {
  Check(buffer); 
  gl_putboxmask(x,y,w,h,buffer); 
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// screen (aka display driver)
// ----------------------------------------------------------------------------
void Screen::init() {
  int err = vga_init();
  if(err) throw EScreen(xprintf("vga_init() fault %d",err));
}
// ----------------------------------------------------------------------------
void Screen::kill() {
  vga_setmode(TEXT);
}
// ----------------------------------------------------------------------------
void Screen::vsync() {
  vga_waitretrace();
}
// ----------------------------------------------------------------------------
void Screen::setRes(int resX, int resY) {
  vga_modeinfo * m; int i;
  for(i=1; i<=vga_lastmodenumber(); i++) 
    if((m = vga_getmodeinfo(i)) && 
        vga_hasmode(i) &&
        m->width==resX && m->height==resY && 
	m->bytesperpixel==1 && m->colors==256) break;
  if(i>vga_lastmodenumber()) 
    throw EScreen(xprintf("non-planar resolution %dx%dx256 not supported "
                          "on this card",resX,resY));
  vga_setmode(i);
  gl_setcontextvga(i);
  
  w=vga_getxdim();
  h=vga_getydim();
  
  gl_enableclipping();
  gl_setclippingwindow(0,0,w-1,h-1);
  
  // get startup palette
  gl_getpalette(palette.data);
  byte * ptr = &palette.data[0].r;
  for(int i=0; i<256*3; i++) *ptr++<<=2;
  
  assert(w==resX);
  assert(h==resY);
  assert(BITSPERPIXEL==8);
}
// ----------------------------------------------------------------------------
// MetaBmp primitives
void Screen::plot(int x, int y, int color) {
  gl_setpixel(x,y,color);
}
// ----------------------------------------------------------------------------
void Screen::draw(int x1, int y1, int x2, int y2, int color) {
  gl_line(x1,y1,x2,y2,color);
}
// ----------------------------------------------------------------------------
void Screen::set(int color) {
  gl_clearscreen(color);
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// KEYBOARD code
// ----------------------------------------------------------------------------
class SVGAKeyboard : public InputDevice {
    class KeyBind {
      public:
        int scancode;
	XStr cmd;
	KeyBind(int s=-1, XStr c="") 
	  : scancode(s), cmd(c) { }
	bool operator == (const KeyBind & r) const 
	  { return scancode==r.scancode; }
    };
    struct transBind { bool special; int woShift, wShift; };
    enum {
      keyUp, keyDown, keyLeft, keyRight,
      keyStrafe, keyStrafeLeft, keyStrafeRight,
      keyShoot,
      keyLast
    };
    
#   define maxKey 120
    bool key[maxKey];
    transBind tt[maxKey]; // tt for translation table
    
    int moveKeys[keyLast];
    Block <KeyBind> binds;
    FIFO <EvKbd*> kbdEvents;
    
    void keyboardEventHandler(int scancode, int press); // called by addThisKEH()
    void consoleHandler(int scancode, int press); // called by addThisConsole()
    friend void addThisKEH(int scancode, int press); 
    friend void addThisConsole(int scancode, int press); 
    
    void clearState();
    void initTT();
  public:
    void init(XStr configFile);
    void turnOn();
    void turnOff();
    void update();
    // console mode
    void eventWaitModeOn() {
      clearState();
      keyboard_seteventhandler(addThisConsole);
    }
    void eventWaitModeOff() {
      clearState();
      keyboard_seteventhandler(addThisKEH);
    }
    void waitForEvent() { while(kbdEvents.isEmpty()) { keyboard_update(); } }
    Event * getEvent() { 
      if(kbdEvents.isEmpty()) return 0; 
      else return kbdEvents.get();
    }
    virtual XStr className() { return "SVGAKeyboard"; }
};

SVGAKeyboard okbd;
InputDevice * kbd = &okbd; // external

// ----------------------------------------------------------------------------
// adds this ptr to event handler ;)
void addThisConsole(int scancode, int press) {
  okbd.consoleHandler(scancode,press);
}
void addThisKEH(int scancode, int press) {
  okbd.keyboardEventHandler(scancode,press);
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::keyboardEventHandler(int scancode, int press) {
  // update key table
  if(scancode>=0 && scancode<sizeofT(key)) key[scancode]=(bool)press;
  // check for ctrl-c
  if( (key[SCANCODE_LEFTCONTROL] || key[SCANCODE_RIGHTCONTROL]) &&
       key[SCANCODE_C] ) haltx("ctrl-c pressed!"); 
  // check if there is a bind for this key
  int i;
  if(press && commandTarget && (i=binds.index(scancode))!=-1)
    commandTarget->handleEvent(new EvCommand(this, binds[i].cmd));
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::clearState() {
  memset(key,0,sizeof(key));
  while(!kbdEvents.isEmpty()) delete kbdEvents.get();
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::turnOn() {
  assert(this==kbd); 
  // other keyboards than kbd are not allowed.
  // (there is only one keyboard device, and what's more importent --
  // ther's only one keyboardEventHandler() func ;)
  
  if(keyboard_init()==-1) 
    throw EDead("keyboard_init() failed. can not initialize keyboard.");
  clearState();
  keyboard_seteventhandler(addThisKEH);
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::turnOff() {
  keyboard_close();
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::init(XStr configFile) {
  // reset settings
  initTT();
  binds.reset();
  moveKeys[keyLeft]        = SCANCODE_CURSORBLOCKLEFT;
  moveKeys[keyRight]       = SCANCODE_CURSORBLOCKRIGHT;
  moveKeys[keyUp]          = SCANCODE_CURSORBLOCKUP;
  moveKeys[keyDown]        = SCANCODE_CURSORBLOCKDOWN;
  moveKeys[keyStrafe]      = SCANCODE_LEFTALT;
  moveKeys[keyStrafeLeft]  = SCANCODE_COMMA;
  moveKeys[keyStrafeRight] = SCANCODE_PERIOD;
  moveKeys[keyShoot]       = SCANCODE_LEFTCONTROL;
  
  FileReader s;
  s.open(configFile);
  while(s.nextClassWord()) {
    XStr word=downcase(s.classWord);
    if(word=="x" || word=="execute") {
      s.nextClassWord();
      int sc=s.classWord.asInt();
      if(sc==0) 
        s.error("scancode can't be 0 or string. was : '"+s.classWord+"'");
      if(!s.nextClassWord()) 
        s.error("x <scancode> must be followed by a command to execute");
      binds.add(KeyBind(sc,s.classWord));
    }
    else if(word=="m" || word=="move") {
      s.nextClassWord();
      int sc=s.classWord.asInt();
      if(sc==0) 
        s.error("scancode can't be 0 or string. was : '"+s.classWord+"'");
      if(!s.nextClassWord()) 
        s.error("m <scancode> must be followed by a movement direction");
      word=downcase(s.classWord);
      
           if(word=="left")        moveKeys[keyLeft]        = sc;
      else if(word=="right")       moveKeys[keyRight]       = sc;
      else if(word=="up")          moveKeys[keyUp]          = sc;
      else if(word=="down")        moveKeys[keyDown]        = sc;
      else if(word=="strafe")      moveKeys[keyStrafe]      = sc;
      else if(word=="strafeleft")  moveKeys[keyStrafeLeft]  = sc;
      else if(word=="straferight") moveKeys[keyStrafeRight] = sc;
      else if(word=="shoot")       moveKeys[keyShoot]       = sc;
      else s.error("unreconized direction : '"+word+"'");
    }
    else s.error(xprintf("unreconaized keyword '%s'",word.chars()));  
  }
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::update() {
  keyboard_update();
  if(moveTarget) {
    EvMove * e = new EvMove(this);
  
    bool s=key[moveKeys[keyStrafe]];
    if(!s && key[moveKeys[keyLeft]])  e->turn=EvMove::turnLeft;
    if(!s && key[moveKeys[keyRight]]) e->turn=EvMove::turnRight;
    if( (s && key[moveKeys[keyLeft]]) || key[moveKeys[keyStrafeLeft]] ) 
      e->dir=EvMove::dirLeft;
    else if( (s && key[moveKeys[keyRight]]) || key[moveKeys[keyStrafeRight]] )
      e->dir=EvMove::dirRight;
    else if(key[moveKeys[keyDown]]) 
      e->dir=EvMove::dirBackward;
    else if(key[moveKeys[keyUp]]) 
      e->dir=EvMove::dirForward;
    
    moveTarget->handleEvent(e);
  }
  
  if(commandTarget && key[moveKeys[keyShoot]]) 
    commandTarget->handleEvent(new EvCommand(this,"shoot"));
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::initTT() {
  struct transBindI {
    int scancode;
    bool special;
    int woShift, wShift;
  };
  static transBindI initTranslationTable[] = {
    { SCANCODE_Q, 0, 'q', 'Q' }, { SCANCODE_W, 0, 'w', 'W' },
    { SCANCODE_E, 0, 'e', 'E' }, { SCANCODE_R, 0, 'r', 'R' },
    { SCANCODE_T, 0, 't', 'T' }, { SCANCODE_Y, 0, 'y', 'Y' },
    { SCANCODE_U, 0, 'u', 'U' }, { SCANCODE_I, 0, 'i', 'I' },
    { SCANCODE_O, 0, 'o', 'O' }, { SCANCODE_P, 0, 'p', 'P' },
    { SCANCODE_A, 0, 'a', 'A' }, { SCANCODE_S, 0, 's', 'S' },
    { SCANCODE_D, 0, 'd', 'D' }, { SCANCODE_F, 0, 'f', 'F' },
    { SCANCODE_G, 0, 'g', 'G' }, { SCANCODE_H, 0, 'h', 'H' },
    { SCANCODE_J, 0, 'j', 'J' }, { SCANCODE_K, 0, 'k', 'K' },
    { SCANCODE_L, 0, 'l', 'L' }, { SCANCODE_Z, 0, 'z', 'Z' },
    { SCANCODE_X, 0, 'x', 'X' }, { SCANCODE_C, 0, 'c', 'C' },
    { SCANCODE_V, 0, 'v', 'V' }, { SCANCODE_B, 0, 'q', 'B' },
    { SCANCODE_N, 0, 'n', 'N' }, { SCANCODE_M, 0, 'm', 'M' },
    
    { SCANCODE_1, 0, '1', '!' }, { SCANCODE_2, 0, '2', '@' },
    { SCANCODE_3, 0, '3', '#' }, { SCANCODE_4, 0, '4', '$' },
    { SCANCODE_5, 0, '5', '%' }, { SCANCODE_6, 0, '6', '^' },
    { SCANCODE_7, 0, '7', '&' }, { SCANCODE_8, 0, '8', '*' },
    { SCANCODE_9, 0, '9', '(' }, { SCANCODE_0, 0, '0', ')' },
    
    { SCANCODE_MINUS,          0, '-',  '_'  },
    { SCANCODE_EQUAL,          0, '=',  '+'  },
    { SCANCODE_BACKSPACE,      0, '\b', '\b' },
    { SCANCODE_TAB,            0, '\t', '\t' },
    { SCANCODE_ESCAPE,         0, '\e', '\e' },
    { SCANCODE_BRACKET_LEFT,   0, '[',  '{'  },
    { SCANCODE_BRACKET_RIGHT,  0, ']',  '}'  },    
    { SCANCODE_SEMICOLON,      0, ';',  ':'  },
    { SCANCODE_APOSTROPHE,     0, '\'', '"'  },
    { SCANCODE_GRAVE,          0, '`',  '~'  },
    { SCANCODE_COMMA,          0, ',',  '<'  },
    { SCANCODE_PERIOD,         0, '.',  '>'  },
    { SCANCODE_SLASH,          0, '/',  '?'  },
    { SCANCODE_SPACE,          0, ' ',  ' '  },
    { SCANCODE_BACKSLASH,      0, '\\', '|'  },
    { SCANCODE_ENTER,          0, '\n', '\n' }, // \n, not \r !
    // keypad
    { SCANCODE_KEYPADPERIOD,   0, '.',  '.'  },
    { SCANCODE_KEYPADMULTIPLY, 0, '*',  '*'  },
    { SCANCODE_KEYPADMINUS,    0, '-',  '-'  },
    { SCANCODE_KEYPADPLUS,     0, '+',  '+'  },
    { SCANCODE_KEYPADENTER,    0, '\n', '\n' },
    { SCANCODE_KEYPADDIVIDE,   0, '/',  '/'  },
    { SCANCODE_KEYPAD1,        0, '1',  '1'  }, 
    { SCANCODE_KEYPAD2,        0, '2',  '2'  },
    { SCANCODE_KEYPAD3,        0, '3',  '3'  }, 
    { SCANCODE_KEYPAD4,        0, '4',  '4'  },
    { SCANCODE_KEYPAD5,        0, '5',  '5'  }, 
    { SCANCODE_KEYPAD6,        0, '6',  '6'  },
    { SCANCODE_KEYPAD7,        0, '7',  '7'  }, 
    { SCANCODE_KEYPAD8,        0, '8',  '8'  },
    { SCANCODE_KEYPAD9,        0, '9',  '9'  }, 
    { SCANCODE_KEYPAD0,        0, '0',  '0'  },
    
    { SCANCODE_CURSORBLOCKUP,    1, EvKbd::up,     EvKbd::up    },
    { SCANCODE_CURSORBLOCKDOWN,  1, EvKbd::down,   EvKbd::down  },
    { SCANCODE_CURSORBLOCKLEFT,  1, EvKbd::left,   EvKbd::left  },
    { SCANCODE_CURSORBLOCKRIGHT, 1, EvKbd::right,  EvKbd::right },
    { SCANCODE_HOME,             1, EvKbd::home,   EvKbd::home   },
    { SCANCODE_END,              1, EvKbd::end,    EvKbd::end    },
    { SCANCODE_PAGEUP,           1, EvKbd::pgUp,   EvKbd::pgUp   },
    { SCANCODE_PAGEDOWN,         1, EvKbd::pgDown, EvKbd::pgDown },
    { SCANCODE_INSERT,           1, EvKbd::ins,    EvKbd::ins    },
    { SCANCODE_REMOVE,           1, EvKbd::del,    EvKbd::del    },
    
    { SCANCODE_F1,  1, EvKbd::f1,  EvKbd::f11 },
    { SCANCODE_F2,  1, EvKbd::f2,  EvKbd::f12 },
    { SCANCODE_F3,  1, EvKbd::f3,  EvKbd::f13 },
    { SCANCODE_F4,  1, EvKbd::f4,  EvKbd::f14 },
    { SCANCODE_F5,  1, EvKbd::f5,  EvKbd::f15 },
    { SCANCODE_F6,  1, EvKbd::f6,  EvKbd::f16 },
    { SCANCODE_F7,  1, EvKbd::f7,  EvKbd::f17 },
    { SCANCODE_F8,  1, EvKbd::f8,  EvKbd::f18 },
    { SCANCODE_F9,  1, EvKbd::f9,  EvKbd::f19 },
    { SCANCODE_F10, 1, EvKbd::f10, EvKbd::f20 },
    { SCANCODE_F11, 1, EvKbd::f11, EvKbd::f11 }, // additional pc f's are
    { SCANCODE_F12, 1, EvKbd::f12, EvKbd::f12 }, // maped to f11,12 regradless
                                                 // of shift state
    { 0, 0, 0, 0 } // terminator
  };
  memset(tt,0,sizeof(tt));
  transBindI * p = initTranslationTable;
  while(p->scancode) {
    assert(p->scancode>0);
    assert(p->scancode<sizeofT(tt));
    tt[p->scancode].special = p->special;  
    tt[p->scancode].wShift  = p->wShift;  
    tt[p->scancode].woShift = p->woShift;  
    p++;
  }
}
// ----------------------------------------------------------------------------
void SVGAKeyboard::consoleHandler(int scancode, int press) {
  // update key table
  if(scancode>=0 && scancode<sizeofT(key)) key[scancode]=(bool)press;
  
  if( (key[SCANCODE_LEFTCONTROL] || key[SCANCODE_RIGHTCONTROL]) &&
       key[SCANCODE_C] ) haltx("ctrl-c pressed!"); 
       
  if(press && scancode>0 && scancode<sizeofT(tt)) {
    int k;
    if(key[SCANCODE_LEFTSHIFT] || key[SCANCODE_RIGHTSHIFT]) 
      k=tt[scancode].wShift;
    else
      k=tt[scancode].woShift;
    if(k)
      kbdEvents.put(new EvKbd(this,tt[scancode].special,k));
  }
}
// ----------------------------------------------------------------------------
void Bmp::lock() {
  assert(buffer); 
  assert(line);
  size_t PAGESIZE = getpagesize();
  lockedMem = malloc( w*h + h*sizeof(byte*) + PAGESIZE-1);
  if(!lockedMem) throw EDead("Bmp::lock(): out of memory!");
  void * p = (void*)(((int)lockedMem + PAGESIZE - 1) & ~(PAGESIZE - 1));
  memcpy(p,buffer,w*h);
  delete [] buffer; buffer=(byte*)p;
  delete [] line;   line=(byte**) ((char*)p+w*h);
  
  byte ** ptr = line; byte * x = buffer; int n=h;
  while(n-->0) { *ptr++=x; x+=w;}
  
  // Mark the buffer read-only. 
  if (mprotect(p, w*h + h*sizeof(char*), PROT_READ)) 
    haltErrno("Couldn't mprotect");
}
// ----------------------------------------------------------------------------
void Bmp::free() {
  if(lockedMem) {
    size_t PAGESIZE = getpagesize();
    if (mprotect((void*)(((int)lockedMem + PAGESIZE - 1) & ~(PAGESIZE - 1)), 
             w*h + h*sizeof(char*), PROT_READ|PROT_WRITE))
      haltErrno("Couldn't un-mprotect");
    ::free(lockedMem);
  }
  else {
    delete [] buffer;
    delete [] line;
  }
  lockedMem=0;
  line=0; buffer=0;
}
// ----------------------------------------------------------------------------


